#include "config.h"
#include "AffixIterator.h"
#include "ContigGraph.h"
#include "ContigPath.h"
#include "DirectedGraphImpl.h"
#include "Estimate.h"
#include "PairUtils.h"
#include "Uncompress.h"
#include <algorithm> // for min
#include <climits> // for UINT_MAX
#include <cmath>
#include <fstream>
#include <getopt.h>
#include <iostream>
#include <iterator>
#include <pthread.h>
#include <sstream>
#include <vector>

using namespace std;

#define PROGRAM "SimpleGraph"

static const char VERSION_MESSAGE[] =
PROGRAM " (" PACKAGE_NAME ") " VERSION "\n"
"Written by Jared Simpson, Shaun Jackman and Jenny Qian.\n"
"\n"
"Copyright 2010 Canada's Michael Smith Genome Science Centre\n";

static const char USAGE_MESSAGE[] =
"Usage: " PROGRAM " [OPTION]... ADJ DIST\n"
"Find paths through contigs using distance estimates.\n"
"  ADJ   adjacency of the contigs\n"
"  DIST  distance estimates between the contigs\n"
"  CONTIG  ids of the contigs\n"
"\n"
"  -k, --kmer=KMER_SIZE  k-mer size\n"
"      --max-cost=COST   maximum computational cost\n"
"  -o, --out=FILE        write result to FILE\n"
"  -c, --cyc-out=FILE        write cyclical paths to FILE\n"
"  -s, --relative-seed-length	seed contig length = 2 x k - relative-seed-length\n"
"  -m, --merging-mode	 extend contigs in both directions from a given contigID file\n"
"  -j, --threads=THREADS use THREADS parallel threads [1]\n"
"  -v, --verbose         display verbose output\n"
"      --help            display this help and exit\n"
"      --version         output version information and exit\n"
"\n"
"Report bugs to <" PACKAGE_BUGREPORT ">.\n";

namespace opt {
	static unsigned k;
	static unsigned maxCost = 100000;
	static unsigned threads = 1;
	static int mode = 0;
	static int s;
	static int verbose;
	static string cycOut;
	static string out;
}

static const char shortopts[] = "j:k:o:c:s:v:m";

enum { OPT_HELP = 1, OPT_VERSION, OPT_MAX_COST };

static const struct option longopts[] = {
	{ "kmer",        required_argument, NULL, 'k' },
	{ "max-cost",    required_argument, NULL, OPT_MAX_COST },
	{ "out",         required_argument, NULL, 'o' },
	{ "cyc-out",     required_argument, NULL, 'c' },
	{ "relative-seed-length", 	required_argument, NULL, 's'},
	{ "threads",     required_argument,	NULL, 'j' },
	{ "verbose",     no_argument, NULL, 'v' },
	{ "merging-mode",	 no_argument, NULL, 'm' },  
	{ "help",        no_argument, NULL, OPT_HELP },
	{ "version",     no_argument, NULL, OPT_VERSION },
	{ NULL, 0, NULL, 0 }
};

typedef vector<LinearNumKey> LinearNumVector;
typedef vector<LinearNumVector > LinearNumKeyVector;

struct SimpleDataCost
{
	SimpleDataCost(size_t kmer) : m_overlap(kmer - 1) { }
	
	size_t cost(const SimpleContigData& data) { return data.length - m_overlap; } 
	size_t m_overlap;
};

static void generatePathsThroughEstimates(
		SimpleContigGraph* pContigGraph, string estFileName);

static void generatePathsThroughUniqueEdges(
                SimpleContigGraph* pContigGraph);

static void generateImplicitPaths(SimpleContigGraph* pContigGraph);

static void generateImplicitPaths(SimpleContigGraph* pContigGraph, std::map<string, bool>& s_contigIDs);

static void constructContigPath(
		const SimpleContigGraph::VertexPath& vertexPath,
		ContigPath& contigPath);
static void scaffoldMultiplePath(
		ostream& vout, SimpleContigGraph::FeasiblePaths& solutions, const SimpleContigGraph* pContigGraph, SimpleDataCost& costFunctor, SimpleContigGraph::FeasiblePaths::iterator& bestSol, ContigPath& ppath, ContigPath& spath, size_t& numN);

static void loadContigIDFromFile(std::map<string, bool>& s_contigIDs, const string& contigFile); 

int main(int argc, char** argv)
{
	bool die = false;
	for (int c; (c = getopt_long(argc, argv,
					shortopts, longopts, NULL)) != -1;) {
		istringstream arg(optarg != NULL ? optarg : "");
		switch (c) {
			case '?': die = true; break;
			case 'j': arg >> opt::threads; break;
			case 'k': arg >> opt::k; break;
			case OPT_MAX_COST: arg >> opt::maxCost; break;
			case 'o': arg >> opt::out; break;

			case 'c': arg >> opt::cycOut; break;
		
			case 's': arg >> opt::s; break;

			case 'm': opt::mode++; break;

			case 'v': opt::verbose++; break;
			case OPT_HELP:
				cout << USAGE_MESSAGE;
				exit(EXIT_SUCCESS);
			case OPT_VERSION:
				cout << VERSION_MESSAGE;
				exit(EXIT_SUCCESS);
		}
	}
	// cout << "v=" << opt::verbose << " k=" << opt::k << " j=" << opt::threads << " m=" << opt::mode << '\n';
	if (opt::k <= 0) {
		cerr << PROGRAM ": missing -k,--kmer option\n";
		die = true;
	}

	if (opt::out.empty()) {
		cerr << PROGRAM ": " << "missing -o,--out option\n";
		die = true;
	}

	//if (argc - optind < 2) {
	//	cerr << PROGRAM ": missing arguments\n";
	//	die = true;

	if (argc - optind < 1) {
		cerr << PROGRAM ": missing arguments\n";
		die = true;
	} else if (argc - optind > 2) {
		cerr << PROGRAM ": too many arguments\n";
		die = true;
	}

	if (die) {
		cerr << "Try `" << PROGRAM
			<< " --help' for more information.\n";
		exit(EXIT_FAILURE);
	}

	string adjFile(argv[optind++]);
	// Load the graph from the adjacency file
        SimpleContigGraph contigGraph;
        loadGraphFromAdjFile(&contigGraph, adjFile);
        if (opt::verbose > 0)
		cerr << "Vertices: " << contigGraph.getNumVertices()
                        << " Edges: " << contigGraph.countEdges() << endl;
	if (argc - optind == 0) {
		// cerr << "Branch 1";
		// try to find paths that are supported by unambiguous adjacency graph
                generatePathsThroughUniqueEdges(&contigGraph);
	}
	if ((argc - optind == 1) && (opt::mode == 0)) {
		// cerr << "Branch 2";
		string estFile(argv[optind++]);

		// try to find paths that match the distance estimates
		generatePathsThroughEstimates(&contigGraph, estFile);
	}
	if ((argc - optind == 1) && (opt::mode > 0)) {
		// cerr << "Branch 3";
		string contigFile(argv[optind++]);

		// Load the contig IDs from the contigID file 
                //std::vector<unsigned> s_contigIDs;
		std::map<string, bool> s_contigIDs;
		loadContigIDFromFile(s_contigIDs, contigFile);

                //unsigned contigIDToIndex(const ContigID& id)
                //{
                //        m_contigDict.push_back(id);
                //        return m_contigDict.size() - 1;
                //}


		// try to extend contigs in the given contigID file
		generateImplicitPaths(&contigGraph, s_contigIDs);
	} 
	if (opt::s > 0) {
		cerr << "Branch 4";
		int seed_length = 2 * opt::k - opt::s;
		cerr << opt::s << endl;
		cerr << seed_length << endl;
		generateImplicitPaths(&contigGraph);
	}
}

template<typename K, typename D>
ostream& printConstraints(ostream& out, const map<K,D>& s)
{
	for (typename map<K,D>::const_iterator iter = s.begin();
			iter != s.end(); ++iter)
		out << ' ' << g_contigIDs.key(iter->first) << iter->second;
	return out;
}

//template<typename K>
//ostream& printPath(ostream& out, const vector<K>& s) 
//{
        //assert(!s.empty());
        //copy(s.begin(), s.end()-1,
                        //ostream_iterator<K>(out, " "));
        //return out << s.back();
//}

template<typename K>
ostream& printPath(ostream& out, const vector<K>& s, size_t first=0, size_t last=0)
{
	if (first+last==s.size()) return out;
	assert(first + last< s.size());
	copy(s.begin() + first, s.end() - (last +1), ostream_iterator<K>(out, " "));
	return out << *(s.end() - (last +1));
}

/** Return the set of contigs that appear more than once in a single
 * solution.
 */
static set<LinearNumKey> findRepeats(
	const SimpleContigGraph::FeasiblePaths& solutions)
{
	set<LinearNumKey> repeats;
	for (SimpleContigGraph::FeasiblePaths::const_iterator
			solIt = solutions.begin();
			solIt != solutions.end(); ++solIt) {
		map<LinearNumKey, unsigned> count;
		for (SimpleContigGraph::VertexPath::const_iterator
				it = solIt->begin(); it != solIt->end(); ++it)
			count[it->key]++;
		for (map<LinearNumKey, unsigned>::const_iterator
				it = count.begin(); it != count.end(); ++it)
			if (it->second > 1)
				repeats.insert(it->first);
	}
	return repeats;
}

/** The fewest number of pairs in a distance estimate. */
static unsigned g_minNumPairs = UINT_MAX;

/** The fewest number of pairs used in a path. */
static unsigned g_minNumPairsUsed = UINT_MAX;

static struct {
	unsigned totalAttempted;
	unsigned noPossiblePaths;
	unsigned nopathEnd;
	unsigned uniqueEnd;
	unsigned multiEnd;
} stats;

/** Return the contig ID of the specified contig serial number. */
static const string& contigNumberToID(LinearNumKey serial)
{
	return g_contigIDs.key(serial);
}

/** Find a path for the specified distance estimates.
 * @return a pointer to the statistic to increment.
 */
static void handleEstimate(
		const EstimateRecord& er, unsigned dirIdx,
		const SimpleContigGraph* pContigGraph,
		// ContigPath& out)
		ContigPath& ppath, size_t& numN, ContigPath& spath)
{
	if (er.estimates[dirIdx].empty())
		return;

	ostringstream vout_ss;
	ostream bitBucket(NULL);
	ostream& vout = opt::verbose > 0 ? vout_ss : bitBucket;
	vout << "\n* " << g_contigIDs.key(er.refID)
		<< (dirIdx ? '-' : '+') << '\n';

	unsigned minNumPairs = UINT_MAX;
	SimpleDataCost costFunctor(opt::k);
	// generate the reachable set
	SimpleContigGraph::KeyConstraintMap constraintMap;
	for (EstimateVector::const_iterator iter
				= er.estimates[dirIdx].begin();
			iter != er.estimates[dirIdx].end(); ++iter) {
		minNumPairs = min(minNumPairs, iter->numPairs);

		// Translate the distances produced by the esimator into the
		// coordinate space used by the graph (a translation of
		// m_overlap)
		int translatedDistance = iter->distance
			+ costFunctor.m_overlap;
		unsigned distanceBuffer = allowedError(iter->stdDev);

		Constraint nc;
		nc.distance = translatedDistance  + distanceBuffer;
		nc.isRC = iter->contig.sense();
		constraintMap[iter->contig.id()] = nc;
	}

	vout << "Constraints:";
	printConstraints(vout, constraintMap) << '\n';

	// Generate the paths
	SimpleContigGraph::FeasiblePaths solutions;

	// The value at which to abort trying to find a path
	const int maxNumPaths = 200;
	int numVisited = 0;

	// The number of nodes to explore before bailing out
	pContigGraph->findSuperpaths(er.refID, (extDirection)dirIdx,
			constraintMap, solutions, costFunctor, maxNumPaths,
			opt::maxCost, numVisited);

	set<LinearNumKey> repeats = findRepeats(solutions);
	if (!repeats.empty()) {
		vout << "Repeats:";
		transform(repeats.begin(), repeats.end(),
				affix_ostream_iterator<string>(vout, " "),
				contigNumberToID);
		vout << '\n';
	}

	unsigned numPossiblePaths = solutions.size();
	if (numPossiblePaths > 0)
		vout << "Paths: " << numPossiblePaths << '\n';

	//size_t max_len = 0;
	for (SimpleContigGraph::FeasiblePaths::iterator solIter
				= solutions.begin(); solIter != solutions.end();) {
		printPath(vout, *solIter) << '\n';

		// Calculate the path distance to each node and see if
		// it is within the estimated distance.
		map<LinearNumKey, int> distanceMap;
		pContigGraph->makeDistanceMap(*solIter,
				costFunctor, distanceMap);

		// Remove solutions whose distance estimates are not correct.
		unsigned validCount = 0, invalidCount = 0, ignoredCount = 0;
		for (EstimateVector::const_iterator iter
					= er.estimates[dirIdx].begin();
				iter != er.estimates[dirIdx].end(); ++iter) {
			vout << *iter << '\t';

			map<LinearNumKey, int>::iterator dmIter
				= distanceMap.find(iter->contig.id());
			assert(dmIter != distanceMap.end());
			// translate distance by -overlap to match
			// coordinate space used by the estimate
			int actualDistance
				= dmIter->second - costFunctor.m_overlap;
			int diff = actualDistance - iter->distance;
			unsigned buffer = allowedError(iter->stdDev);
			bool invalid = (unsigned)abs(diff) > buffer;
			bool repeat = repeats.count(iter->contig.id()) > 0;
			bool ignored = invalid && repeat;
			if (ignored)
				ignoredCount++;
			else if (invalid)
				invalidCount++;
			else
				validCount++;
			vout << "dist: " << actualDistance
				<< " diff: " << diff
				<< " buffer: " << buffer
				<< " n: " << iter->numPairs
				<< (ignored ? " ignored" : invalid ? " invalid" : "")
				<< '\n';
		}

		if (invalidCount == 0 && validCount > 0)
				++solIter;
		else
			solIter = solutions.erase(solIter);
	}

	vout << "Solutions: " << solutions.size() << '\n';

	SimpleContigGraph::FeasiblePaths::iterator bestSol
		= solutions.end();
	int minDiff = 999999;
	for (SimpleContigGraph::FeasiblePaths::iterator solIter
				= solutions.begin();
			solIter != solutions.end(); ++solIter) {
		map<LinearNumKey, int> distanceMap;
		pContigGraph->makeDistanceMap(*solIter,
				costFunctor, distanceMap);
		int sumDiff = 0;
		for (EstimateVector::const_iterator iter
					= er.estimates[dirIdx].begin();
				iter != er.estimates[dirIdx].end(); ++iter) {
			if (repeats.count(iter->contig.id()) > 0)
				continue;
			map<LinearNumKey, int>::iterator dmIter
				= distanceMap.find(iter->contig.id());
			if (dmIter != distanceMap.end()) {
				int actualDistance = dmIter->second
					- costFunctor.m_overlap;
				int diff = actualDistance - iter->distance;
				sumDiff += abs(diff);
			}
		}

		if (sumDiff < minDiff) {
			minDiff = sumDiff;
			bestSol = solIter;
		}

		size_t len = pContigGraph->calculatePathLength(
				*solIter, costFunctor);
		printPath(vout, *solIter)
			<< " length: " << len
			<< " sumdiff: " << sumDiff << '\n';
	}

	/** Lock the debugging stream. */
	static pthread_mutex_t coutMutex = PTHREAD_MUTEX_INITIALIZER;
	pthread_mutex_lock(&coutMutex);
	stats.totalAttempted++;
	g_minNumPairs = min(g_minNumPairs, minNumPairs);
	// cout << vout_ss.str();

	if (numPossiblePaths == 0) {
		stats.noPossiblePaths++;
	} else if (solutions.empty()) {
		stats.nopathEnd++;
	} else if (solutions.size() > 1) {
		// vout << "Scaffolding here... " << '\n';	
		// Scaffolding code here.
		stats.multiEnd++;
		scaffoldMultiplePath(vout, solutions, pContigGraph, costFunctor, bestSol, ppath, spath, numN);
		g_minNumPairsUsed = min(g_minNumPairsUsed, minNumPairs);
	} else {
		assert(solutions.size() == 1);
		assert(bestSol != solutions.end());
		// constructContigPath(*bestSol, out);
		constructContigPath(*bestSol, ppath);
		numN = 0;
		spath.clear();
		stats.uniqueEnd++;
		g_minNumPairsUsed = min(g_minNumPairsUsed, minNumPairs);
	}
	cout << vout_ss.str();
	pthread_mutex_unlock(&coutMutex);
}

/** Load contigIDs from a file.
 *  @return a vector.
 */

static void loadContigIDFromFile(std::map<string, bool>& s_contigIDs, const string& contigFile)
{

	ifstream in(contigFile.c_str());
	assert(in.is_open());
	
	unsigned count = 0;
	string id;
	while (in >> id) {
		in.ignore(numeric_limits<streamsize>::max(), '\n');
		//s_contigIDs.push_back(id);
		s_contigIDs[id] = true;
		if (++count % 1000000 == 0)
			cout << "Read " << count << " ids" << endl;
	}
	assert(in.eof());
	in.close();
}
		

/** Find a path for unambiguous adjacency graph.
 * @return a pointer to the statistic to increment.
 */
static void handleEdges(
		ostream& vout,
                const LinearNumKey& id, unsigned dirIdx,
                const SimpleContigGraph* pContigGraph,
                ContigPath& out,
		SeenVertexTable& s_vertexTable,
		bool& isCircular)
{
    vout << "\n* " << g_contigIDs.key(id)
                << (dirIdx ? '-' : '+') << '\n';

        // Generate the path
	SimpleContigGraph::VertexPath path;
    
    SimpleContigGraph::VertexType* pSourceVertex = pContigGraph->findVertex(id);
    // Create the initial (empty) path
    pContigGraph->UnambiguousDFS(pSourceVertex, (extDirection)dirIdx, false, path, s_vertexTable, isCircular);
	
	// print out the path to vout
	if (path.size() >= 1) {
               	printPath(vout, path) << '\n';
		constructContigPath (path, out);
	}

}

/** Find a path for unambiguous adjacency graph.
 * @return a pointer to the statistic to increment.
 */
static void handleImplicitPaths(
                ostream& vout,
                const LinearNumKey& id, unsigned dirIdx,
                const SimpleContigGraph* pContigGraph,
                ContigPath& out,
                SeenVertexTable& s_vertexTable,
                bool& isCircular)
{
    vout << "\n* " << g_contigIDs.key(id)
                << (dirIdx ? '-' : '+') << '\n';

        // Generate the path
        SimpleContigGraph::VertexPath path;

    SimpleContigGraph::VertexType* pSourceVertex = pContigGraph->findVertex(id);
    
    // Create a SeenVertexTable for the path
    SeenVertexTable path_s_vertexTable;
    // Create the initial (empty) path
    pContigGraph->ImplicitDFS(pSourceVertex, (extDirection)dirIdx, false, path, s_vertexTable, isCircular, path_s_vertexTable);

        // print out the path to vout
        if (path.size() >= 1) {
                printPath(vout, path) << '\n';
                constructContigPath (path, out);
        }

}



struct WorkerArg {
	istream* in;
	ostream* out;
	const SimpleContigGraph* graph;
	WorkerArg(istream* in, ostream* out, const SimpleContigGraph* g)
		: in(in), out(out), graph(g) { }
};

struct WorkerEArg {
	ostream* out;

	ostream* out_c;

	const SimpleContigGraph* graph;
	//WorkerEArg(ostream* out, const SimpleContigGraph* g)
	//	: out(out), graph(g) { }

	WorkerEArg(ostream* out, ostream* out_c, const SimpleContigGraph* g)
		: out(out), out_c(out_c), graph(g) { 
		//*out_c << "Hello World";
		}
};

struct WorkerIArg {
	ostream* out;
	ostream* out_c;
	const SimpleContigGraph* graph;
	std::map<string, bool>& s_contigIDs;
	WorkerIArg(ostream* out, ostream* out_c, const SimpleContigGraph* g, std::map<string, bool>& s_contigIDs)
		: out(out), out_c(out_c), graph(g), s_contigIDs(s_contigIDs) { }
};

static void* workerE (void* pArg)
{
        WorkerEArg& arg = *static_cast<WorkerEArg*>(pArg);

	SeenVertexTable s_vertexTable;

        ostringstream vout_ss;
        ostream bitBucket(NULL);
        ostream& vout = opt::verbose > 0 ? vout_ss : bitBucket;
	LinearNumKey nid;
        for (nid = 0; nid < arg.graph->getNumVertices(); nid++) {
		//vout << "* " << g_contigIDs.key(nid) << '\n';
		// Early exit and skip this vertex if the sourceKey has already been visited
    		if (s_vertexTable[nid])
        		continue;
   
                ContigPath path;

		//ContigPath path_c;

		path.push_back(MergeNode(nid, false));

		bool isCircular = false;
                //handleEdges(vout, nid, ANTISENSE, arg.graph, path, s_vertexTable);
		handleEdges(vout, nid, ANTISENSE, arg.graph, path, s_vertexTable, isCircular);
		if (!isCircular) {
		reverse(path.begin(), path.end());
		//handleEdges(vout, nid, SENSE, arg.graph, path, s_vertexTable);
		handleEdges(vout, nid, SENSE, arg.graph, path, s_vertexTable, isCircular);
		}
		if (path.size() > 1) {
                        /** Lock the output stream. */
                        static pthread_mutex_t outMutex
                                = PTHREAD_MUTEX_INITIALIZER;
                        pthread_mutex_lock(&outMutex);
                        //*arg.out << g_contigIDs.key(nid) << '\t' << '\t' << '\t' << path << ' ';
			if (isCircular)
				//*arg.out_c << g_contigIDs.key(nid) << '\t' << path.size() << '\t' << path << '\n';
				*arg.out_c << g_contigIDs.key(nid) << '\t' << path << '\n';
			else
				*arg.out << g_contigIDs.key(nid) << '\t' << path << '\n';
                        assert(arg.out->good());
			assert(arg.out_c->good());
                        pthread_mutex_unlock(&outMutex);
                }
        }
	/** Lock the debugging stream. */
	static pthread_mutex_t coutMutex = PTHREAD_MUTEX_INITIALIZER;
	pthread_mutex_lock(&coutMutex);
	cout << vout_ss.str();
	pthread_mutex_unlock(&coutMutex);
        return NULL;
}

static void* workerI (void* pArg)
{
        WorkerIArg& arg = *static_cast<WorkerIArg*>(pArg);

        SeenVertexTable s_vertexTable;

        ostringstream vout_ss;
        ostream bitBucket(NULL);
        ostream& vout = opt::verbose > 0 ? vout_ss : bitBucket;
        LinearNumKey nid;
        for (nid = 0; nid < arg.graph->getNumVertices(); nid++) {
                //vout << "* " << g_contigIDs.key(nid) << '\n';
                // Early exit and skip this vertex if the sourceKey has already been visited
                if (s_vertexTable[nid])
                        continue;

                //SimpleContigGraph::VertexType* pSourceVertex = pContigGraph->findVertex(id);

                //SimpleContigGraph::VertexType* pSourceVertex = arg.graph->findVertex(nid);
		string ID = g_contigIDs.key(nid);

                //int seedContigLength = 2 * opt::k -opt::s;
                //if (pSourceVertex->m_data.length != seedContigLength)
                //        continue;

		if (!arg.s_contigIDs[ID])
			continue;
 
                ContigPath path;

                //ContigPath path_c;

                path.push_back(MergeNode(nid, false));
                // testing code for circular paths
                // if (nid == 207860)
                //    *arg.out_c << path << '\n';

                bool isCircular = false;
                //handleEdges(vout, nid, ANTISENSE, arg.graph, path, s_vertexTable);
                handleImplicitPaths(vout, nid, ANTISENSE, arg.graph, path, s_vertexTable, isCircular);
                if (isCircular) {
                    //*arg.out_c << path << '\n';
                    reverse(path.begin(), path.end());
                    //*arg.out_c << path << '\n';
                    handleImplicitPaths(vout, nid, SENSE, arg.graph, path, s_vertexTable, isCircular);
                    //*arg.out_c << path << '\n';
                        }
                if (!isCircular) {
                reverse(path.begin(), path.end());
                //handleEdges(vout, nid, SENSE, arg.graph, path, s_vertexTable);
                handleImplicitPaths(vout, nid, SENSE, arg.graph, path, s_vertexTable, isCircular);
                }
                if (path.size() > 1) {
                        /** Lock the output stream. */
                        static pthread_mutex_t outMutex
                                = PTHREAD_MUTEX_INITIALIZER;
                        pthread_mutex_lock(&outMutex);
                        //*arg.out << g_contigIDs.key(nid) << '\t' << '\t' << '\t' << path << ' ';
                        if (isCircular)
                                //*arg.out_c << g_contigIDs.key(nid) << '\t' << path.size() << '\t' << path << '\n';
                                *arg.out_c << g_contigIDs.key(nid) << '\t' << path << '\n';
                        else
                                *arg.out << g_contigIDs.key(nid) << '\t' << path << '\n';
                        assert(arg.out->good());
                        assert(arg.out_c->good());
                        pthread_mutex_unlock(&outMutex);
                }
        }
        /** Lock the debugging stream. */
        static pthread_mutex_t coutMutex = PTHREAD_MUTEX_INITIALIZER;
        pthread_mutex_lock(&coutMutex);
        cout << vout_ss.str();
        pthread_mutex_unlock(&coutMutex);
        return NULL;
}

static void* workerIl (void* pArg)
{
        WorkerEArg& arg = *static_cast<WorkerEArg*>(pArg);

        SeenVertexTable s_vertexTable;

        ostringstream vout_ss;
        ostream bitBucket(NULL);
        ostream& vout = opt::verbose > 0 ? vout_ss : bitBucket;
        LinearNumKey nid;
        for (nid = 0; nid < arg.graph->getNumVertices(); nid++) {
                //vout << "* " << g_contigIDs.key(nid) << '\n';
                // Early exit and skip this vertex if the sourceKey has already been visited
                if (s_vertexTable[nid])
                        continue;

		//SimpleContigGraph::VertexType* pSourceVertex = pContigGraph->findVertex(id);
		
		SimpleContigGraph::VertexType* pSourceVertex = arg.graph->findVertex(nid);
		
		int seedContigLength = 2 * opt::k -opt::s;
		if (pSourceVertex->m_data.length != seedContigLength)
			continue;
  
                ContigPath path;

                //ContigPath path_c;

                path.push_back(MergeNode(nid, false));
		// testing code for circular paths
		// if (nid == 207860)
		//    *arg.out_c << path << '\n';

                bool isCircular = false;
                //handleEdges(vout, nid, ANTISENSE, arg.graph, path, s_vertexTable);
                handleImplicitPaths(vout, nid, ANTISENSE, arg.graph, path, s_vertexTable, isCircular);
                if (isCircular) {
		    *arg.out_c << path << '\n';
		    reverse(path.begin(), path.end());
		    *arg.out_c << path << '\n';
		    handleImplicitPaths(vout, nid, SENSE, arg.graph, path, s_vertexTable, isCircular);
		    *arg.out_c << path << '\n';
			}
		if (!isCircular) {
                reverse(path.begin(), path.end());
                //handleEdges(vout, nid, SENSE, arg.graph, path, s_vertexTable);
                handleImplicitPaths(vout, nid, SENSE, arg.graph, path, s_vertexTable, isCircular);
                }
                if (path.size() > 1) {
                        /** Lock the output stream. */
                        static pthread_mutex_t outMutex
                                = PTHREAD_MUTEX_INITIALIZER;
                        pthread_mutex_lock(&outMutex);
                        //*arg.out << g_contigIDs.key(nid) << '\t' << '\t' << '\t' << path << ' ';
                        if (isCircular)
                                //*arg.out_c << g_contigIDs.key(nid) << '\t' << path.size() << '\t' << path << '\n';
                                *arg.out_c << g_contigIDs.key(nid) << '\t' << path << '\n';
                        else
                                *arg.out << g_contigIDs.key(nid) << '\t' << path << '\n';
                        assert(arg.out->good());
                        assert(arg.out_c->good());
                        pthread_mutex_unlock(&outMutex);
                }
        }
        /** Lock the debugging stream. */
        static pthread_mutex_t coutMutex = PTHREAD_MUTEX_INITIALIZER;
        pthread_mutex_lock(&coutMutex);
        cout << vout_ss.str();
        pthread_mutex_unlock(&coutMutex);
        return NULL;
}


static void* worker(void* pArg)
{
	WorkerArg& arg = *static_cast<WorkerArg*>(pArg);
	for (;;) {
		/** Lock the input stream. */
		static pthread_mutex_t inMutex = PTHREAD_MUTEX_INITIALIZER;
		pthread_mutex_lock(&inMutex);
		EstimateRecord er;
		bool good = (*arg.in) >> er;
		pthread_mutex_unlock(&inMutex);
		if (!good)
			break;

		// ContigPath path;
		// handleEstimate(er, ANTISENSE, arg.graph, path);
		// reverse(path.begin(), path.end());
		ContigPath rppath, rspath, ppath, spath;
                size_t rnumN, numN;
                handleEstimate(er, ANTISENSE, arg.graph, rppath, rnumN, rspath);
		reverse(rppath.begin(), rppath.end());
		reverse(rspath.begin(), rspath.end());
                
		// path.push_back(MergeNode(er.refID, false));
		handleEstimate(er, SENSE, arg.graph, ppath, numN, spath);
		// if (path.size() > 1) {
		if (rppath.size() + rnumN + rspath.size() + ppath.size() + numN + spath.size() > 0) {
			/** Lock the output stream. */
			static pthread_mutex_t outMutex
				= PTHREAD_MUTEX_INITIALIZER;
			pthread_mutex_lock(&outMutex);
			*arg.out << g_contigIDs.key(er.refID) << '\t';
				// << path << '\n';
			if (rspath.size() > 0)
				*arg.out << rspath << ' ';
			if (rnumN > 0)
				*arg.out << rnumN << "N ";
			if (rppath.size() > 0)
				*arg.out << rppath << ' ';
			*arg.out << MergeNode(er.refID, false);
			if (ppath.size() > 0)
				*arg.out << ' ' << ppath;
			if (numN > 0)
				*arg.out << ' ' << numN << 'N';
			if (spath.size() > 0)
				*arg.out << ' ' << spath;
			*arg.out << '\n';
			assert(arg.out->good());
			pthread_mutex_unlock(&outMutex);
		}
	}
	exit(0);
	// return NULL;
}

static void generatePathsThroughEstimates(
		SimpleContigGraph* pContigGraph, string estFileName)
{
	ifstream inStream(estFileName.c_str());
	ofstream outStream(opt::out.c_str());
	assert(outStream.is_open());

	// Create the worker threads.
	vector<pthread_t> threads;
	threads.reserve(opt::threads);
	WorkerArg arg(&inStream, &outStream, pContigGraph);
	for (unsigned i = 0; i < opt::threads; i++) {
		pthread_t thread;
		pthread_create(&thread, NULL, worker, &arg);
		threads.push_back(thread);
	}

	// Wait for the worker threads to finish.
	for (vector<pthread_t>::const_iterator it = threads.begin();
			it != threads.end(); ++it) {
		void* status;
		pthread_join(*it, &status);
	}

	cout <<
		"Total paths attempted: " << stats.totalAttempted << "\n"
		"No possible paths: " << stats.noPossiblePaths << "\n"
		"No valid paths: " << stats.nopathEnd << "\n"
		"Multiple valid paths: " << stats.multiEnd << "\n"
		"Unique path: " << stats.uniqueEnd << "\n";

	inStream.close();
	outStream.close();

	cout << "\n"
		"The minimum number of pairs in a distance estimate is "
		<< g_minNumPairs << ".\n";
	if (g_minNumPairsUsed != UINT_MAX) {
		cout << "The minimum number of pairs used in a path is "
			<< g_minNumPairsUsed << ".\n";
		if (g_minNumPairs < g_minNumPairsUsed)
			cout << "Consider increasing the number of pairs "
				"threshold paramter, n, to " << g_minNumPairsUsed
				<< ".\n";
	}
}

static void generatePathsThroughUniqueEdges(
                SimpleContigGraph* pContigGraph)
{
        ofstream outStream(opt::out.c_str());
        assert(outStream.is_open());

	ofstream outStream_c(opt::cycOut.c_str());
	assert(outStream_c.is_open());

        // Create the worker threads.
        vector<pthread_t> threads;
        threads.reserve(opt::threads);
        //WorkerEArg arg(&outStream, pContigGraph);

	WorkerEArg arg(&outStream, &outStream_c, pContigGraph);

        for (unsigned i = 0; i < opt::threads; i++) {
                pthread_t thread;
                pthread_create(&thread, NULL, workerE, &arg);
                threads.push_back(thread);
        }

        // Wait for the worker threads to finish.
        for (vector<pthread_t>::const_iterator it = threads.begin();
                        it != threads.end(); ++it) {
                void* status;
                pthread_join(*it, &status);
        }
        outStream.close();
	outStream_c.close();
}

static void generateImplicitPaths(
                SimpleContigGraph* pContigGraph, std::map<string, bool>& s_contigID)
{
	ofstream outStream(opt::out.c_str());
        assert(outStream.is_open());

        ofstream outStream_c(opt::cycOut.c_str());
        assert(outStream_c.is_open());

        // Create the worker threads.
        vector<pthread_t> threads;
        threads.reserve(opt::threads);

        WorkerIArg arg(&outStream, &outStream_c, pContigGraph, s_contigID);

        for (unsigned i = 0; i < opt::threads; i++) {
                pthread_t thread;
                pthread_create(&thread, NULL, workerI, &arg);
                threads.push_back(thread);
        }

        // Wait for the worker threads to finish.
        for (vector<pthread_t>::const_iterator it = threads.begin();
                        it != threads.end(); ++it) {
                void* status;
                pthread_join(*it, &status);
        }
        outStream.close();
        outStream_c.close();
}

static void generateImplicitPaths(
                SimpleContigGraph* pContigGraph)
{
        ofstream outStream(opt::out.c_str());
        assert(outStream.is_open());

        ofstream outStream_c(opt::cycOut.c_str());
        assert(outStream_c.is_open());

        // Create the worker threads.
        vector<pthread_t> threads;
        threads.reserve(opt::threads);

        WorkerEArg arg(&outStream, &outStream_c, pContigGraph);

        for (unsigned i = 0; i < opt::threads; i++) {
                pthread_t thread;
                pthread_create(&thread, NULL, workerIl, &arg);
                threads.push_back(thread);
        }

        // Wait for the worker threads to finish.
        for (vector<pthread_t>::const_iterator it = threads.begin();
                        it != threads.end(); ++it) {
                void* status;
                pthread_join(*it, &status);
        }
        outStream.close();
        outStream_c.close();
}


//
// Convert the vertext path to a contig path
// The difference between the two is that the complementness of a vertex path is with respect to the previous node in the path
// but with a contig path it is with respect to the root contig. Also the contigPath uses true contig ids
//
static void constructContigPath(
		const SimpleContigGraph::VertexPath& vertexPath,
		ContigPath& contigPath)
{
	bool flip = false;
	for(SimpleContigGraph::VertexPath::const_iterator iter = vertexPath.begin(); iter != vertexPath.end(); ++iter)
	{
		flip = flip ^ iter->isRC;
		contigPath.push_back(MergeNode(iter->key, flip));
	}
}

static void scaffoldMultiplePath(ostream& vout, SimpleContigGraph::FeasiblePaths& solutions, const SimpleContigGraph* pContigGraph, SimpleDataCost& costFunctor, SimpleContigGraph::FeasiblePaths::iterator& bestSol, ContigPath& ppath, ContigPath& spath, size_t& numN)
{
		SimpleContigGraph::FeasiblePaths::iterator solIter = solutions.begin();
		SimpleContigGraph::VertexPath& firstSol = *solIter;
		size_t min_len = firstSol.size();
                while (++solIter != solutions.end()) {
			if (solIter->size()<min_len)
				min_len = solIter->size();
		}
		// get the index for longestPrefix
		SimpleContigGraph::VertexPath vppath;
		size_t longestPrefix;
		bool commonPrefix = true;
		for (longestPrefix = 0; longestPrefix < min_len; longestPrefix++) {
			// Get element in ith position in first solution:  firstSol[longestPrefix]
			const SimpleContigGraph::PathNode& common_path_node = firstSol[longestPrefix];
			// Iterate over all solutions.
			for (SimpleContigGraph::FeasiblePaths::iterator solIter = solutions.begin(); solIter != solutions.end();++solIter) {
				// Get element in ith position in next solution:  (*solIter)[longestPrefix]
				const SimpleContigGraph::PathNode& pathnode = (*solIter)[longestPrefix];
				// If not matched: longestPrefix is the length of the longest prefix, must break out of the outer and inner loop
				if (pathnode.key != common_path_node.key || pathnode.isRC != common_path_node.isRC) { 
					commonPrefix = false;
					break;
				}
			}
			if (!commonPrefix)
				break;
			vppath.push_back(common_path_node);

		}

		// get the index for longestSuffix
		SimpleContigGraph::VertexPath vspath;
		size_t longestSuffix;
		bool commonSuffix = true;
		for (longestSuffix = 0; longestSuffix < min_len-longestPrefix; longestSuffix++) {
			// Get element in ith position in first solution:  firstSol[firstSol.size()-longestSuffix-1]
			SimpleContigGraph::PathNode& common_path_node = firstSol[firstSol.size()-longestSuffix-1];
			// Iterate over all solutions.
			for (SimpleContigGraph::FeasiblePaths::iterator solIter = solutions.begin(); solIter != solutions.end();++solIter) {
				// Compare with element in ith position in next solution:  (*solIter)[solIter->size()-longestSuffix-1]
				SimpleContigGraph::PathNode& pathnode = (*solIter)[solIter->size()-longestSuffix-1];
				// If not matched: longestSuffix is the length of the longest suffix, must break out of the outer and inner loop
				if (pathnode.key != common_path_node.key || pathnode.isRC != common_path_node.isRC) { 
					commonSuffix = false;
			        	break; // must break out of the outer and inner loop
				}
			}
			if (!commonSuffix) {
				break;
			}
			vspath.push_back(common_path_node);
		}
		reverse(vspath.begin(), vspath.end());

		// print out the longest prefix and the longest suffix
		vout << "Longest Prefix: " << '\n';	
		printPath(vout, firstSol,0,firstSol.size()-longestPrefix);
		vout << '\n' << "Longest Suffix: " << '\n';
		printPath(vout, firstSol, firstSol.size()-longestSuffix, 0);
		vout << '\n';
		// find the longest path
		size_t max_len = 0;
		for (SimpleContigGraph::FeasiblePaths::iterator solIter
                                = solutions.begin(); solIter != solutions.end();) {
                	// add only paths whose length is maximal
                	size_t len = pContigGraph->calculatePathLength(
                      		*solIter, costFunctor);
			// vout << "Solutions: " << '\n';
			// for (SimpleContigGraph::FeasiblePaths::iterator prtIter
                        //        = solutions.begin(); prtIter != solutions.end();prtIter++) 
			//	printPath(vout, *prtIter) << '\n';	
			// vout << "Path length: " << len << '\n';
                	if (len >= max_len) {
                      		if (len > max_len) {
                                	// erase previous non-maximal paths
                              		solIter = solutions.erase(solutions.begin(), solIter);
					bestSol = solIter;
					max_len = len;
				}
                      		++solIter;
                	}
                	else {
				// vout << "Erasing Solution: " << '\n';
				// printPath(vout, *solIter) << '\n';	
                      		solIter = solutions.erase(solIter);
			}
	       	}	
			// vout << "Solutions: " << '\n';
			// for (SimpleContigGraph::FeasiblePaths::iterator prtIter
                        //        = solutions.begin(); prtIter != solutions.end();prtIter++) 
			//	printPath(vout, *prtIter) << '\n';	
		//stats.multiEnd++;
		// if (solutions.size() != 1)
			// cout << vout_ss.str();
		// assert(solutions.size() == 1);
		bestSol = solutions.begin();
			
                //constructContigPath(*bestSol, out);
		ppath.clear();
                constructContigPath(vppath, ppath);
		spath.clear();
                constructContigPath(vspath, spath);
		numN = pContigGraph->calculatePathLength(*bestSol, costFunctor, longestPrefix, longestSuffix);
		//// subBestSol = *bestSol[0, *bestSol.
		//// constructPartialContigPath(*prefix, Ns, *suffix, out); 
                //stats.uniqueEnd++;
                //g_minNumPairsUsed = min(g_minNumPairsUsed, minNumPairs);
}
